#!/usr/bin/env python3
# PYTHON_ARGCOMPLETE_OK
# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

from __future__ import annotations

import argparse
import json
import os
import shutil
import sys
from collections.abc import Collection, Iterable, Iterator, MutableSequence, Sequence
from contextlib import ExitStack, contextmanager
from pathlib import Path
from tempfile import TemporaryDirectory
from textwrap import dedent
from typing import IO, Any, NamedTuple, cast

from license_expression import ExpressionError
from zstarfile import ZSTarfile

from go_vendor_tools import __version__
from go_vendor_tools.archive import get_toplevel_directory
from go_vendor_tools.cli.utils import (
    HAS_TOMLKIT,
    catch_vendor_tools_error,
    load_tomlkit_if_exists,
    need_tomlkit,
    tomlkit_dump,
)
from go_vendor_tools.config.base import load_config
from go_vendor_tools.config.licenses import (
    LicenseConfig,
    LicenseEntry,
    create_license_config,
)
from go_vendor_tools.exceptions import VendorToolsError
from go_vendor_tools.gomod import get_unlicensed_mods
from go_vendor_tools.hashing import get_hash
from go_vendor_tools.license_detection.base import LicenseData, LicenseDetector
from go_vendor_tools.license_detection.load import DETECTORS, get_detectors
from go_vendor_tools.licensing import compare_licenses, simplify_license
from go_vendor_tools.specfile import VendorSpecfile

if HAS_TOMLKIT:
    import tomlkit

try:
    import argcomplete
except ImportError:
    HAS_ARGCOMPLETE = False
else:
    HAS_ARGCOMPLETE = True

COLOR: bool | None = None
RED = "\033[31m"  # ]
CLEAR = "\033[0m"  # ]

MANUALLY_DETECTING_LICENSES_URL = "https://fedora.gitlab.io/sigs/go/go-vendor-tools/scenarios/#manually-detecting-licenses"


def red(__msg: str, /, *, file: IO[str] | None = None) -> None:
    file = cast(IO[str], sys.stdout if file is None else file)
    color = COLOR
    if color is None:
        color = file.isatty()
    print(f"{RED if color else ''}{__msg}{CLEAR if color else ''}", file=file)


def split_kv_options(kv_config: list[str]) -> dict[str, str]:
    results: dict[str, str] = {}
    for opt in kv_config:
        if ";" in opt:
            results |= split_kv_options(opt.split(";"))
        else:
            key, _, value = opt.partition("=")
            results[key] = value
    return results


def choose_license_detector(
    choice: str | None,
    license_config: LicenseConfig,
    kv_config: list[str] | None,
    find_only: bool = False,
) -> LicenseDetector:
    kv_config = kv_config or []
    cli_config = license_config["detector_config"] | split_kv_options(kv_config)
    available, missing = get_detectors(cli_config, license_config, find_only=find_only)
    if choice:
        if choice in missing:
            sys.exit(f"Failed to get detector {choice!r}: {missing[choice]}")
        return available[choice]
    if not available:
        print("Failed to load license detectors:", file=sys.stderr)
        for detector, err in missing.items():
            print(f"! {detector}: {err}")
        sys.exit(1)
    return next(iter(available.values()))


def _add_json_argument(parser: argparse.ArgumentParser, **kwargs) -> None:
    our_kwargs: dict[str, Any] = {
        "type": Path,
        "help": dedent(
            """\
        Write license data to a JSON file.
        This data is not yet considered stable and is only intended for
        go2rpm's internal usage and for testing purposes.
        """
        ).replace("\n", " "),
    }
    kwargs = our_kwargs | kwargs
    parser.add_argument("--write-json", **kwargs)


def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Handle licenses for vendored go projects",
        prog="go_vendor_license",
    )
    parser.add_argument("--version", action="version", version=__version__)
    parser.add_argument("-c", "--config", type=Path, dest="config_path")
    parser.add_argument(
        "-C",
        "--path",
        "--directory",
        type=Path,
        dest="directory",
        action="append",
        help=dedent(
            """
        Can be one of the following:

        1. Top-level directory with a go.mod file and vendor directory
        2. If --use-archive is specified, treat as a tarball and unpack it. Can
           be specified multiple times to unpack multiple tarballs on
           top of one another.
        3. Path to a specfile.
           The paths to Source0 and Source1 will be automatically unpacked.
        """
        ),
    )
    parser.add_argument("--use-archive", action="store_true", help="See --path.")
    parser.add_argument(
        "--color",
        action=argparse.BooleanOptionalAction,
        default=False if os.environ.get("NO_COLOR") else None,
        help="Whether to use colored output."
        " Defaults to True if output is a TTY and $NO_COLOR is not defined.",
    )
    parser.add_argument(
        "-d",
        "--detector",
        choices=DETECTORS,
        default=None,
        help="Choose a license detector. Choices: %(choices)s. Default: autodetect",
        dest="detector_name",
    )
    parser.add_argument(
        "-D",
        "--dc",
        "--detector-config",
        help="`KEY=VALUE` pairs to pass to the license detector."
        " Can be passed multiple times."
        " Overrides settings defined in licensing.detector_config.",
        dest="detector_config",
        action="append",
    )
    parser.set_defaults(detector_find_only=False)
    subparsers = parser.add_subparsers(dest="subcommand")
    subparsers.required = True
    report_parser = subparsers.add_parser(
        "report",
        help="Main subcommand",
        description="This command detects licenses within the project tree."
        " It creates a license summary and a normalized SPDX expression",
    )
    report_parser.add_argument(
        "-i",
        "--ignore-undetected",
        action="store_true",
        help="Whether to show undetected licenses in the output",
    )
    report_parser.add_argument(
        "-L",
        "--ignore-unlicensed-mods",
        action="store_true",
        help="Whether to show Go modules without licenses in the output",
    )
    report_parser.add_argument(
        "--subpackage-name",
        help="""
        Which subpackage to use to find license tag.
        Name of a subpackage.

        - `-NAME` for `%%package NAME`;
        - `NAME` for `%%package -n NAME`;
        - Omit this args to use the main package definition
        """,
    )
    verify_opts = report_parser.add_mutually_exclusive_group()
    verify_opts.add_argument(
        "--verify",
        help="Verify license expression to make sure it matches caluclated expression",
        metavar="EXPRESSION",
    )
    verify_opts.add_argument(
        "--verify-spec", help="Verify specfile license tag", action="store_true"
    )
    verify_opts.add_argument(
        "--update-spec", help="Update specfile license tag", action="store_true"
    )
    report_parser.add_argument(
        "--prompt",
        action=argparse.BooleanOptionalAction,
        help="Whether to prompt to fill in undetected licenses."
        " Implies --write-config."
        " Default: %(default)s",
    )
    report_parser.add_argument(
        "mode",
        nargs="?",
        type=str,
        choices=("all", "expression", "list"),
        default="all",
        help="""
        - `all` — print out a breakdown of all license files and their detected
          license expression and then a final, cummluative expression.
        - `expression` — print only the cummulative SPDX expression
        - `list` — print the file-by-file breakdown only
        """,
    )
    _add_json_argument(report_parser)
    report_parser.add_argument(
        "--write-config", help="Write a base config.", action="store_true"
    )
    help_msg = "Add manual license entry to a config file"
    explict_parser = subparsers.add_parser(
        "explicit",
        help=help_msg,
        description=f"{help_msg}. See {MANUALLY_DETECTING_LICENSES_URL} for usage.",
    )
    explict_parser.add_argument(
        "-f",
        "--file",
        dest="license_file",
        required=True,
        type=Path,
        help="Path to file (relative to CWD) to add to license config",
    )
    explict_parser.add_argument("license_expression", help="SPDX license expression")
    install_parser = subparsers.add_parser(
        "install", description=f"INTERNAL: {install_command.__doc__}"
    )
    install_parser.add_argument(
        "--install-directory", dest="install_directory", type=Path, required=True
    )
    install_parser.add_argument(
        "--destdir", dest="install_destdir", type=Path, default=Path("/")
    )
    install_parser.add_argument(
        "--filelist", dest="install_filelist", type=Path, required=True
    )
    install_parser.set_defaults(detector_find_only=True)
    # TODO(gotmax23): Should we support writing JSON from the install command
    # or just reading it?
    # _add_json_argument(install_parser)
    generate_buildrequires_parser = subparsers.add_parser(
        "generate_buildrequires",
        description="Internal command for %%go_vendor_license_buildrequires",
    )
    generate_buildrequires_parser.add_argument(
        "--no-check",
        help="Whether to exclude dependencies for %%go_vendor_license_check",
        action="store_true",
        dest="detector_find_only",
    )
    return parser


def parseargs(argv: list[str] | None = None) -> argparse.Namespace:
    """
    Parse arguments and return an `argparse.Namespace`
    """
    parser = get_parser()
    if HAS_ARGCOMPLETE:
        argcomplete.autocomplete(parser)
    args = parser.parse_args(argv)
    args.directory = list(map(Path, args.directory or (".")))
    if args.subcommand not in ("explicit",):
        loaded = load_config(
            args.config_path, allow_missing=getattr(args, "write_config", False)
        )
        args.config = loaded["licensing"]
        if not args.detector_name:
            args.detector_name = args.config["detector"]
    if args.subcommand in ("report", "install"):
        args.detector = choose_license_detector(
            args.detector_name,
            args.config,
            args.detector_config,
            args.detector_find_only,
        )
        # TODO(anyone): Replace the print if/when we implement more granular logging
        print("Using detector:", args.detector.NAME, file=sys.stderr)
    global COLOR  # noqa: PLW0603
    COLOR = args.color
    return args


def bullet_iterator(it: Iterable[object], bullet: str = "- ") -> Iterator[str]:
    for item in it:
        yield bullet + str(item)


def red_if_true(items: Collection[object], message: str, bullet: str = "- ") -> None:
    if not items:
        return
    print(message)
    red("\n".join(bullet_iterator(items, bullet)))


def paths_relative_to_list(paths: Collection[Path], directory: Path) -> list[Path]:
    return [
        path.resolve().relative_to(directory.resolve()) if path.is_absolute() else path
        for path in paths
    ]


def print_licenses(
    results: LicenseData,
    unlicensed_mods: Collection[Path],
    mode: str,
    show_undetected: bool,
    show_unlicensed: bool,
    directory: Path,
) -> None:
    if mode in ("all", "list"):
        for (
            license_path,
            license_name,
        ) in results.license_map.items():
            print(f"{license_path}: {license_name}")
    if (
        results.undetected_licenses
        or unlicensed_mods
        or results.unmatched_extra_licenses
    ):
        if mode != "expression":
            print()
        if show_undetected:
            red_if_true(
                results.undetected_licenses,
                "The following license files were found "
                "but the correct license identifier couldn't be determined:",
            )
        if show_unlicensed:
            red_if_true(
                paths_relative_to_list(unlicensed_mods, directory),
                "The following modules are missing license files:",
            )
        red_if_true(
            results.unmatched_extra_licenses,
            "The following license files that were specified in the configuration"
            " have changed:",
        )
    if mode == "list":
        return
    if mode != "expression":
        print()
    print(results.license_expression)


def write_license_json(data: LicenseData, file: Path) -> None:
    with file.open("w", encoding="utf-8") as fp:
        json.dump(data.to_jsonable(), fp, indent=2)


class _PromptMissingResult(NamedTuple):
    data: LicenseData
    entries: MutableSequence[LicenseEntry]
    excludes: list[str]


# TODO(gotmax23): Unit test prompt_missing_licenses and write_config code.
# This'll require some mocking of the input() stuff.
def prompt_missing_licenses(
    data: LicenseData,
    entries: MutableSequence[LicenseEntry],
) -> _PromptMissingResult:
    excludes: list[str] = []
    if not data.undetected_licenses:
        return _PromptMissingResult(data, entries, excludes)
    print("Undetected licenses found! Please enter them manually.")
    undetected_licenses = set(data.undetected_licenses)
    license_map: dict[Path, str] = dict(data.license_map)
    for undetected in sorted(data.undetected_licenses):
        print(f"* Undetected license: {data.directory / undetected}")
        expression_str = input("Enter SPDX expression (or EXCLUDE): ")
        if expression_str == "EXCLUDE":
            undetected_licenses.remove(undetected)
            excludes.append(str(undetected))
            print("Adding file to licensing.exclude_files...")
            continue
        expression: str = (
            str(simplify_license(expression_str)) if expression_str else ""
        )
        print(f"Expression simplified to {expression!r}")
        license_map[undetected] = expression
        entry_dict = LicenseEntry(
            path=str(undetected),
            sha256sum=get_hash(data.directory / undetected),
            expression=expression,
        )
        replace_entry(entries, entry_dict, undetected)
        undetected_licenses.remove(undetected)
    assert not undetected_licenses
    return _PromptMissingResult(
        data.replace(undetected_licenses=undetected_licenses, license_map=license_map),
        entries,
        excludes,
    )


def _write_config_verify_path(config_path: Path | None) -> Path:
    if config_path:
        return config_path
    need_tomlkit("--write-config")

    default = Path.cwd() / "go-vendor-tools.toml"
    if not default.is_file():
        sys.exit("--write-config: Please pass --config to write configuration file!")
    else:
        print(
            "WARNING --write-config: No --config path specified"
            f" Will write to {default}",
            file=sys.stderr,
        )
    return config_path or default


def get_report_write_config_data(
    config_path: Path | None, detector: LicenseDetector
) -> tuple[Path, tomlkit.TOMLDocument]:
    need_tomlkit("--write-config")
    new_config_path = _write_config_verify_path(config_path)
    loaded = load_tomlkit_if_exists(config_path)
    write_config_data = loaded.setdefault("licensing", {})
    write_config_data["detector"] = detector.NAME
    if detector.detector_config:
        write_config_data["detector_config"] = detector.detector_config
    return new_config_path, loaded


def write_and_prompt_report_licenses(
    license_data: LicenseData, write_config_data: tomlkit.TOMLDocument
) -> LicenseData:
    # fmt: off
    license_config_list = (
        write_config_data
        .setdefault("licensing", {})
        .setdefault("licenses", tomlkit.aot() if HAS_TOMLKIT else [])
    )
    # fmt: on
    license_data, _, exclude_files = prompt_missing_licenses(
        license_data, license_config_list
    )
    if exclude_files:
        exclude_files_toml = write_config_data["licensing"].setdefault(  # type: ignore[union-attr]
            "exclude_files", []
        )
        exclude_files_toml.extend(exclude_files)
    return license_data


@contextmanager
def handle_alternative_sources_and_spec(
    directories: Sequence[Path], is_archive: bool, subpackage_name: str | None
) -> Iterator[tuple[Path, VendorSpecfile | None]]:
    with ExitStack() as es:
        spec: VendorSpecfile | None = None
        directories = list(directories)
        # The CLI code already checks that the value is greater than zero, so
        # assert is fine here
        assert directories
        if not is_archive and len(directories) > 1:
            sys.exit("Too many paths were passed!")

        if directories[0].suffix == ".spec":
            spec = es.enter_context(VendorSpecfile(directories[0], subpackage_name))
            directories[:] = spec.source0_and_source1()
            is_archive = True
        if is_archive:
            tmp = Path(es.enter_context(TemporaryDirectory()))
            # Extract the first archive
            with ZSTarfile.open(directories[0]) as tar:
                first_toplevel = get_toplevel_directory(tar)
                if not first_toplevel:
                    sys.exit(f"{directories[0]} does not have a top-level directory!")
                print(f"Extracting {directories[0]}", file=sys.stderr)
                tar.extractall(tmp)
            for directory in directories[1:]:
                with ZSTarfile.open(directory) as tar:
                    toplevel = get_toplevel_directory(tar)
                    print(f"Extracting {directory}", file=sys.stderr)
                    tar.extractall(tmp if toplevel else tmp / first_toplevel)
            yield tmp / first_toplevel, spec

        else:
            yield directories[0], spec


def report_command(args: argparse.Namespace) -> None:
    detector: LicenseDetector = args.detector
    paths: Sequence[Path] = args.directory
    ignore_undetected: bool = args.ignore_undetected
    ignore_unlicensed_mods: bool = args.ignore_unlicensed_mods
    mode: str = args.mode
    verify: str | None = args.verify
    write_json: Path | None = args.write_json
    write_config: bool = args.write_config or args.prompt
    prompt: bool = args.prompt
    config_path: Path | None = args.config_path
    use_archive: bool = args.use_archive
    subpackage_name: str | None = args.subpackage_name
    verify_spec: bool = args.verify_spec
    update_spec: bool = args.update_spec
    del args

    if write_config:
        config_path, loaded = get_report_write_config_data(config_path, detector)

    with handle_alternative_sources_and_spec(paths, use_archive, subpackage_name) as (
        directory,
        spec,
    ):
        if (verify_spec or update_spec) and not spec:
            raise VendorToolsError(
                "--path must be a path to a specfile"
                " if --verify-spec or --update-spec is passed"
            )
        license_data: LicenseData = detector.detect(directory)
        unlicensed_mods = (
            set()
            if ignore_unlicensed_mods
            else get_unlicensed_mods(directory, license_data.license_file_paths)
        )
        if prompt:
            license_data = write_and_prompt_report_licenses(license_data, loaded)
        failed = bool(
            (license_data.undetected_licenses and not ignore_undetected)
            or (unlicensed_mods and not ignore_unlicensed_mods)
            or license_data.unmatched_extra_licenses
        )
        print_licenses(
            license_data,
            unlicensed_mods,
            mode,
            not ignore_undetected,
            not ignore_unlicensed_mods,
            directory,
        )
        if write_json:
            write_license_json(license_data, write_json)
        if spec and verify_spec:
            verify = spec.license
        if (
            spec
            and update_spec
            and not compare_licenses(
                spec.license, exp := license_data.license_expression
            )
        ):
            if failed:
                print(
                    "Did not update specfile license tag due to above detector errors."
                )
            else:
                spec.license = str(exp)
        if write_config:
            tomlkit_dump(loaded, cast(Path, config_path))
    if verify and not compare_licenses(license_data.license_expression, verify):
        raise VendorToolsError("Failed to verify license. Expected ^")
    sys.exit(failed)


def _get_intermediate_directories(
    directory_parts: Collection[Sequence[str]],
) -> set[Path]:
    inter_parts = set()
    for parts in directory_parts:
        for i in range(len(parts)):
            inter_parts.add(Path(*parts[: i + 1]))
    return inter_parts


def copy_licenses(
    base_directory: Path,
    license_paths: Iterable[Path],
    install_destdir: Path,
    install_directory: Path,
    install_filelist: Path,
) -> None:
    installdir = install_destdir / install_directory.relative_to("/")
    base_directory = base_directory.resolve()

    entries: list[str] = []
    directory_parts: set[Sequence[str]] = set()
    installdir.mkdir(parents=True, exist_ok=True)
    entries.append(f"%license %dir {install_directory}")
    for lic in license_paths:
        resolvedpath = lic.resolve()
        relpath = resolvedpath.relative_to(base_directory)
        if len(relpath.parts) > 1:
            directory_parts.add(relpath.parts[:-1])
        (installdir / relpath).parent.mkdir(parents=True, exist_ok=True)
        shutil.copy2(lic, installdir / relpath)
        entries.append(f"%license {install_directory / relpath}")
    entries.extend(
        f"%license %dir {install_directory / path}"
        for path in _get_intermediate_directories(directory_parts)
    )
    entries.sort()
    install_filelist.write_text("\n".join(entries) + "\n")


def install_command(args: argparse.Namespace) -> None:
    """
    Install license files into the license directory
    """
    directory: Path = args.directory[0]
    detector: LicenseDetector = args.detector
    install_destdir: Path = args.install_destdir
    install_directory: Path = args.install_directory
    install_filelist: Path = args.install_filelist
    del args

    license_files = detector.find_license_files(directory)
    copy_licenses(
        directory,
        license_files,
        install_destdir,
        install_directory,
        install_filelist,
    )


def get_relpath(base_directory: Path, path: Path) -> Path:
    if path.is_absolute():
        return path.relative_to(base_directory)
    return path


def replace_entry(
    data: MutableSequence[LicenseEntry], new_entry: LicenseEntry, relpath: Path
) -> None:
    for entry in data:
        if entry == new_entry:
            return
        if Path(entry["path"]) == relpath:
            cast(dict, entry).clear()
            entry.update(new_entry)
            return
    data.append(new_entry)


def explicit_command(args: argparse.Namespace) -> None:
    if not args.config_path:
        sys.exit("--config must be specified!")
    loaded = load_tomlkit_if_exists(args.config_path)

    if "licensing" not in loaded:
        loaded.add("licensing", tomlkit.table())
    data = loaded["licensing"]

    licenses = cast(dict, data).setdefault("licenses", tomlkit.aot())
    relpath = get_relpath(args.directory[0], args.license_file)
    try:
        expression = (
            simplify_license(args.license_expression) if args.license_expression else ""
        )
    except ExpressionError as exc:
        sys.exit(f"Failed to parse license: {exc}")
    entry = LicenseEntry(
        path=str(relpath),
        sha256sum=get_hash(args.license_file),
        expression=expression,
    )
    replace_entry(licenses, entry, relpath)
    tomlkit_dump(loaded, args.config_path)


def generate_buildrequires_command(args: argparse.Namespace) -> None:
    detector: str = args.detector_name
    find_only: bool = args.detector_find_only
    del args

    if not detector:
        # If the detector is not explicitly specified, attempt to fall back to
        # the one whose dependencies are already installed.
        available, missing = get_detectors({}, create_license_config())
        detector = next(iter(available), "") or next(iter(missing))
    elif detector not in DETECTORS:
        sys.exit(f"{detector!r} does not exist! Choices: {tuple(DETECTORS)}")
    detector_cls = DETECTORS[detector]
    for requirement in (
        detector_cls.FIND_PACKAGES_NEEDED if find_only else detector_cls.PACKAGES_NEEDED
    ):
        print(requirement)


def main(argv: list[str] | None = None) -> None:
    args = parseargs(argv)
    with catch_vendor_tools_error():
        if args.subcommand == "report":
            report_command(args)
        elif args.subcommand == "explicit":
            explicit_command(args)
        elif args.subcommand == "install":
            install_command(args)
        elif args.subcommand == "generate_buildrequires":
            generate_buildrequires_command(args)


if __name__ == "__main__":
    main()



# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

"""
Utilities for working with license expressions
"""

from __future__ import annotations

from collections.abc import Callable
from functools import lru_cache, partial
from typing import Any, cast

import license_expression
from boolean.boolean import DualBase


def get_fedora_licensing() -> license_expression.Licensing:
    """
    Get a Licensing object modifed to prefer the LicenseRef-Fedora-* extensions
    """
    license_index = license_expression.get_license_index()
    public_dict: dict[str, Any] = {
        "key": "LicenseRef-Fedora-Public-Domain",
        "aliases": [],
        "is_exception": False,
    }
    lics: list[dict[str, Any]] = [public_dict]
    for lic in license_index:
        if lic.get("is_deprecated", False) or not lic.get("spdx_license_key"):
            continue
        ld = {
            "key": lic.get("spdx_license_key", ""),
            "aliases": lic.get("other_spdx_license_keys", []),
            "is_exception": lic.get("is_exception", ""),
        }
        if ld["key"].startswith("LicenseRef-scancode-public-"):
            public_dict["aliases"].append(ld["key"])
            public_dict["aliases"].extend(
                a for a in ld["aliases"] if a != public_dict["key"]
            )
            continue
        lics.append(ld)
    return license_expression.load_licensing_from_license_index(lics)


licensing = get_fedora_licensing()


def combine_licenses(
    *expressions: str | license_expression.LicenseExpression | None,
    validate=True,
    strict=True,
    recursive_simplify: bool = False,
) -> str:
    """
    Combine SPDX license expressions with AND
    """
    converter = cast(
        "Callable[[str | license_expression.LicenseExpression], str]",
        (
            partial(simplify_license, validate=False, strict=False)
            if recursive_simplify
            else str
        ),
    )
    # Set a file's license to an empty string or None to exclude it from the
    # calculation.
    filtered = [converter(expression) for expression in expressions if expression]
    filtered.sort()
    return simplify_license(
        str(license_expression.combine_expressions(filtered, licensing=licensing)),
        validate=validate,
        strict=strict,
    )


def _sort_expression_recursive(
    parsed: license_expression.LicenseExpression, /
) -> license_expression.LicenseExpression:
    if isinstance(parsed, DualBase) and (args := getattr(parsed, "args", None)):
        rec_sorted = sorted((_sort_expression_recursive(arg) for arg in args))
        parsed = parsed.__class__(*rec_sorted)
    return parsed


@lru_cache(500)
def _parse(
    expression: str | license_expression.LicenseExpression,
    validate: bool = True,
    strict: bool = True,
) -> license_expression.LicenseExpression:
    return licensing.parse(str(expression), validate=validate, strict=strict)


@lru_cache(500)
def simplify_license(
    expression: str | license_expression.LicenseExpression,
    *,
    validate: bool = True,
    strict: bool = True,
) -> str:
    """
    Simplify and verify a license expression
    """
    parsed = _parse(expression, validate=validate, strict=strict)
    # DualBase subclasses are collections of licenses with an "AND" or an "OR"
    # relationship.
    if not isinstance(parsed, DualBase):
        return str(parsed)
    # Flatten licenses (e.g., "(MIT AND ISC) AND MIT" -> "MIT AND ISC"
    parsed = parsed.flatten()
    # Perform further license_expression-specific deduplication
    parsed = licensing.dedup(parsed)
    # Recursively sort AND/OR expressions
    parsed = _sort_expression_recursive(parsed)
    return str(parsed)


def validate_license(expression: str) -> bool:
    try:
        _parse(expression)
    except license_expression.ExpressionError:
        return False
    else:
        return True


def compare_licenses(
    license1: license_expression.LicenseExpression | str,
    license2: str | license_expression.LicenseExpression | str,
    /,
    allow_invalid: bool = True,
) -> bool:
    try:
        return simplify_license(license1, validate=False) == simplify_license(
            license2, validate=False
        )
    except license_expression.ExpressionError:
        if not allow_invalid:
            raise
    return False



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

from __future__ import annotations

import sys
from pathlib import Path
from tempfile import TemporaryDirectory

from mkdocs_gen_files.editor import FilesEditor
from releaserr.scd import scd2md

sys.path.insert(0, str(Path(__file__).parent))
from go_vendor_license_help_to_md import get_lines

HERE = Path(__file__).resolve().parent

editor = FilesEditor.current()


def main() -> None:
    files = list(HERE.glob("*.scd"))
    with TemporaryDirectory() as temp:
        # mandir = Path(editor.directory, "man")
        # mandir.mkdir()
        new_files: list[Path] = scd2md(files, Path(temp))
        for file in new_files:
            with editor.open(f"man/{file.name}", "w") as fp:
                fp.write(file.read_text())
    with editor.open("man/go_vendor_license.md", "w") as fp:
        fp.write("\n".join(get_lines(True)))


main()



# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

from __future__ import annotations

import json
import sys
from io import StringIO
from pathlib import Path
from shutil import copy2
from subprocess import CalledProcessError
from textwrap import dedent
from typing import Any

import pytest
from pytest_mock import MockerFixture

from go_vendor_tools.cli import go_vendor_license, utils
from go_vendor_tools.config.base import load_config
from go_vendor_tools.exceptions import MissingDependencyError
from go_vendor_tools.license_detection.askalono import AskalonoLicenseDetector
from go_vendor_tools.license_detection.base import (
    LicenseData,
    LicenseDetector,
    LicenseDetectorNotAvailableError,
    get_manual_license_entries,
)
from go_vendor_tools.license_detection.load import DETECTORS
from go_vendor_tools.license_detection.load import get_detectors as gd

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib

HERE = Path(__file__).resolve().parent
TEST_DATA = HERE / "test_data"

CONFIG1 = load_config(TEST_DATA / "case1" / "config.toml")
CONFIG1_BROKEN = load_config(TEST_DATA / "case1" / "config-broken.toml")


def get_available_detectors() -> list[type[LicenseDetector]]:
    # TODO(anyone): Allow enforcing "strict mode" if any detectors are missing
    # This can be a env var and then enabled in the noxfile.
    available, missing = gd({}, CONFIG1["licensing"])
    # HACK: We initialize the classes using a test config to check if they are
    # available and then return the base class so that it can be reinitialized
    return [type(d) for d in available.values()]


@pytest.fixture(name="detector", params=get_available_detectors())
def get_detectors(request) -> type[LicenseDetector]:
    return request.param


def test_license_explicit(test_data: Path, tmp_path: Path) -> None:
    case_dir = test_data / "case1"
    licenses_dir = case_dir / "licenses"
    with open(case_dir / "config.toml", "rb") as fp:
        expected = tomllib.load(fp)
    dest = tmp_path / "config.toml"
    copy2(case_dir / "config-broken.toml", dest)
    go_vendor_license.main(
        [
            f"-c{dest}",
            f"-C{licenses_dir}",
            "explicit",
            f"-f{licenses_dir / 'LICENSE.MIT'}",
            "MIT",
        ]
    )
    with open(dest, "rb") as fp:
        gotten = tomllib.load(fp)
    assert gotten == expected


def test_get_extra_licenses(test_data: Path) -> None:
    case_dir = test_data / "case1"
    licenses_dir = case_dir / "licenses"
    config = load_config(case_dir / "config.toml")
    matched, missing = get_manual_license_entries(
        config["licensing"]["licenses"], licenses_dir
    )
    expected_map = {
        Path("LICENSE.BSD3"): "BSD-3-Clause",
        Path("LICENSE.MIT"): "MIT",
    }
    assert matched == expected_map
    assert not missing


def test_get_extra_licenses_error(test_data: Path) -> None:
    case_dir = test_data / "case1"
    licenses_dir = case_dir / "licenses"
    matched, missing = get_manual_license_entries(
        CONFIG1_BROKEN["licensing"]["licenses"], licenses_dir
    )
    expected_map = {Path("LICENSE.BSD3"): "BSD-3-Clause"}
    assert matched == expected_map
    assert missing == [Path("LICENSE.MIT")]


@pytest.mark.parametrize(
    "case_name, allowed_detectors, cli_config",
    [
        pytest.param("case2", None, {}, id="case2"),
        pytest.param("case3", [AskalonoLicenseDetector], {"multiple": "1"}, id="case3"),
    ],
)
def test_load_dump_license_data(
    test_data: Path,
    detector: type[LicenseDetector],
    case_name: str,
    allowed_detectors: list[type[LicenseDetector]] | None,
    cli_config: dict[str, str],
    mocker: MockerFixture,
) -> None:
    if allowed_detectors is not None and detector not in allowed_detectors:
        pytest.skip(f"{case_name} does use {detector}")

    # Needed for case3
    mocker.patch("go_vendor_tools.gomod.get_go_module_names", return_value={"abc": ""})

    case_dir = test_data / case_name
    expected_report = case_dir / "reports" / f"{detector.NAME}.json"
    licenses_dir = case_dir / "licenses"
    config = load_config(None)
    detector_obj = detector(cli_config, config["licensing"])
    try:
        data: LicenseData = detector_obj.detect(licenses_dir)
    except Exception as exc:
        print(exc)
        if isinstance(exc, CalledProcessError):
            print("stdout:", exc.stdout)
            print("stderr:", exc.stderr)
            if case_name == "case3" and "Found argument '--multiple'" in exc.stderr:
                # stderr: error: Found argument '--multiple' which wasn't
                # expected, or isn't valid in this context
                # For some reason, this only happens on EL 9.
                pytest.xfail()
        raise

    placeholder_path = Path("/placeholder")
    data.license_file_paths = tuple(
        sorted(
            placeholder_path / path.relative_to(data.directory)
            for path in data.license_file_paths
        )
    )
    data.directory = placeholder_path

    jsonable = data.to_jsonable()
    new_data = type(data).from_jsonable(jsonable)
    new_data.license_file_paths = tuple(
        sorted(
            placeholder_path / path.relative_to(data.directory)
            for path in data.license_file_paths
        )
    )
    assert new_data.to_jsonable() == jsonable

    _remove_license_scanner_data(jsonable)
    # (expected_report).write_text(json.dumps(jsonable, indent=2))
    with (expected_report).open() as fp:
        gotten_json = _remove_license_scanner_data(json.load(fp))
    assert gotten_json == jsonable


def _remove_license_scanner_data(data: dict[str, Any]) -> dict[str, Any]:
    """
    Remove license-scanner specific data from license data dict, as this data
    tends to be unstable and change between versions.
    We only care about the data that's actually produced by g-v-t.
    """
    for name in DETECTORS:
        key = f"{name}_license_data"
        data.pop(key, None)
    return data


def test_detect_nothing(tmp_path: Path, detector: type[LicenseDetector]) -> None:
    """
    Ensure the code has proper error handling for when no licenses are detected
    """
    config = load_config(None)
    detector_obj = detector({}, config["licensing"])
    data: LicenseData = detector_obj.detect(tmp_path)
    assert data.directory == tmp_path
    assert not data.license_map
    assert not data.undetected_licenses
    assert not data.license_set
    assert data.license_expression is None


def test_need_tomlkit(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(utils, "HAS_TOMLKIT", False)
    with pytest.raises(
        MissingDependencyError,
        match="tomlkit is required for this action. Please install it!",
    ):
        go_vendor_license.need_tomlkit()


def test_choose_license_detector_error_1(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(
        "go_vendor_tools.license_detection.scancode.HAS_SCANCODE", False
    )
    with pytest.raises(
        SystemExit,
        match="Failed to get detector 'scancode':"
        " The scancode-toolkit library must be installed!",
    ):
        go_vendor_license.choose_license_detector(
            "scancode", CONFIG1["licensing"], None
        )


def test_choose_license_detector_error_2(
    mocker: MockerFixture, capsys: pytest.CaptureFixture
) -> None:
    return_value: tuple[dict, dict] = (
        {},
        {
            "abcd": LicenseDetectorNotAvailableError("acbd is missing!?!?"),
            "123": LicenseDetectorNotAvailableError("123 is missing."),
        },
    )
    gd_mock = mocker.patch(
        "go_vendor_tools.cli.go_vendor_license.get_detectors",
        return_value=return_value,
    )
    with pytest.raises(SystemExit, match="1"):
        go_vendor_license.choose_license_detector(None, CONFIG1["licensing"], None)
    out, err = capsys.readouterr()
    assert err == "Failed to load license detectors:\n"
    expected = """\
    ! abcd: acbd is missing!?!?
    ! 123: 123 is missing.
    """
    assert dedent(expected) == out
    gd_mock.assert_called_once()


def test_red() -> None:
    with StringIO() as stream:
        go_vendor_license.red("This is an error", file=stream)
        value = stream.getvalue()
    assert value == "This is an error\n"
    with StringIO() as stream:
        stream.isatty = lambda: True  # type: ignore
        go_vendor_license.red("This is an error", file=stream)
        value = stream.getvalue()
    assert value == "\033[31mThis is an error\033[0m\n"


def test_print_licenses_all(capsys: pytest.CaptureFixture) -> None:
    directory = Path("/does-not-exist")
    license_data = LicenseData(
        directory=directory,
        license_map={
            Path("LICENSE.md"): "MIT",
            Path("vendor/xyz/COPYING"): "GPL-3.0-only",
        },
        undetected_licenses=[
            Path("LICENSE.undetected"),
            Path("vendor/123/COPYING.123"),
        ],
        unmatched_extra_licenses=[
            Path("LICENSE-Custom"),
            Path("vendor/custom/LICENSE"),
        ],
        extra_license_files=[],
        detector_name="",
    )
    go_vendor_license.print_licenses(
        results=license_data,
        unlicensed_mods=[
            Path("LICENSE.unmatched"),
            Path("vendor/123/456/LICENSE.unmatched1"),
        ],
        mode="all",
        show_undetected=True,
        show_unlicensed=True,
        directory=directory,
    )
    out, err = capsys.readouterr()
    assert not err
    expected = """\
    LICENSE.md: MIT
    vendor/xyz/COPYING: GPL-3.0-only

    The following license files were found but the correct license identifier couldn't be determined:
    - LICENSE.undetected
    - vendor/123/COPYING.123
    The following modules are missing license files:
    - LICENSE.unmatched
    - vendor/123/456/LICENSE.unmatched1
    The following license files that were specified in the configuration have changed:
    - LICENSE-Custom
    - vendor/custom/LICENSE

    GPL-3.0-only AND MIT
    """  # noqa: E501
    assert out == dedent(expected)


def test_generate_buildrequires(capsys: pytest.CaptureFixture):
    go_vendor_license.main(["--detector=askalono", "generate_buildrequires"])
    out, err = capsys.readouterr()
    assert not err
    assert out == "askalono-cli\n"


def test_generate_buildrequires_no_check(capsys: pytest.CaptureFixture):
    go_vendor_license.main(
        ["--detector=askalono", "generate_buildrequires", "--no-check"]
    )
    out, err = capsys.readouterr()
    assert not err
    assert not out


def test_generate_buildrequires_trivy(capsys: pytest.CaptureFixture):
    go_vendor_license.main(["--detector=trivy", "generate_buildrequires"])
    out, err = capsys.readouterr()
    assert not err
    assert out == "trivy\n"


def test_generate_buildrequires_no_check_trivy(capsys: pytest.CaptureFixture):
    go_vendor_license.main(["--detector=trivy", "generate_buildrequires", "--no-check"])
    out, err = capsys.readouterr()
    assert not err
    assert out == "trivy\n"



# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

"""
scancode-toolkit-based license detector backend
"""

from __future__ import annotations

from collections.abc import Iterable
from dataclasses import dataclass
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING, Any, NamedTuple, TypedDict, cast

from go_vendor_tools.gomod import get_go_module_dirs
from go_vendor_tools.license_detection.search import find_license_files

try:
    import scancode.api  # type: ignore[import]
except ImportError:
    HAS_SCANCODE = False
else:
    HAS_SCANCODE = True

from go_vendor_tools.config.licenses import LicenseConfig
from go_vendor_tools.license_detection.base import (
    LicenseData,
    LicenseDetector,
    LicenseDetectorNotAvailableError,
    get_manual_license_entries,
    reuse_path_to_license_map,
)
from go_vendor_tools.licensing import combine_licenses

if TYPE_CHECKING:
    from _typeshed import StrPath


class ScancodeLicenseDict(TypedDict):
    """
    License data returned by the scancode library
    """

    detected_license_expression: str
    detected_license_expression_spdx: str
    license_detections: list[dict[str, Any]]
    license_clues: list[Any]
    percentage_of_license_text: float


class ScancodeResult(NamedTuple):
    scancode_license_data: dict[str, ScancodeLicenseDict]
    simplified_map: dict[Path, str]
    undetected: set[Path]


def get_scancode_license_data(
    directory: Path,
    files: Iterable[Path],
) -> ScancodeResult:
    data_dicts: dict[str, ScancodeLicenseDict] = {}
    simplified_map: dict[Path, str] = {}
    undetected: set[Path] = set()
    for file in files:
        data = cast(
            ScancodeLicenseDict, scancode.api.get_licenses(str(directory / file))
        )
        data["license_detections"].sort(
            key=lambda d: d.get("license_expression_spdx", "")
        )
        data_dicts[str(file)] = data

        if data["detected_license_expression_spdx"] is None:
            undetected.add(file)
        else:
            simplified_map[file] = data["detected_license_expression_spdx"]
    return ScancodeResult(data_dicts, simplified_map, undetected)


@dataclass()
class ScancodeLicenseData(LicenseData):
    """
    scancode-toolkit-specific LicenseData implementation
    """

    scancode_license_data: dict[str, ScancodeLicenseDict]
    _combine_licenses = staticmethod(partial(combine_licenses, recursive_simplify=True))


class ScancodeLicenseDetector(LicenseDetector[ScancodeLicenseData]):
    NAME = "scancode"
    PACKAGES_NEEDED = ("go-vendor-tools+scancode",)

    def __init__(
        self,
        detector_config: dict[str, str],
        license_config: LicenseConfig,
        find_only: bool = False,
    ) -> None:
        self._find_only = find_only
        if not self.find_only and not HAS_SCANCODE:
            raise LicenseDetectorNotAvailableError(
                "The scancode-toolkit library must be installed!"
            )
        self.detector_config = detector_config
        self.license_config = license_config

    def detect(self, directory: StrPath):
        if self.find_only:
            raise ValueError(
                "This cannot be called when class was initalized with find_only=True"
            )
        directory = Path(directory)
        reuse_roots = get_go_module_dirs(Path(directory), relative_paths=True)
        license_file_lists = find_license_files(
            directory,
            relative_paths=True,
            exclude_directories=self.license_config["exclude_directories"],
            exclude_files=self.license_config["exclude_files"],
            reuse_roots=reuse_roots,
        )
        data, license_map, undetected = get_scancode_license_data(
            directory, map(Path, license_file_lists["license"])
        )
        manual_license_map, manual_unmatched = get_manual_license_entries(
            self.license_config["licenses"], directory
        )
        license_map |= manual_license_map
        license_map |= reuse_path_to_license_map(license_file_lists["reuse"])
        return ScancodeLicenseData(
            directory=directory,
            license_map=license_map,
            undetected_licenses=undetected,
            unmatched_extra_licenses=manual_unmatched,
            scancode_license_data=data,
            # FIXME(gotmax): Change the design of LicenseData to not require full paths
            extra_license_files=[
                Path(directory, file) for file in license_file_lists["notice"]
            ],
            detector_name=self.NAME,
        )



# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

"""
Detect licenses using askalono
"""

from __future__ import annotations

import dataclasses
import json
import shutil
import subprocess
from collections.abc import Callable, Collection
from pathlib import Path
from typing import TYPE_CHECKING, TypedDict, cast

from license_expression import ExpressionError

from go_vendor_tools.config.utils import str_to_bool
from go_vendor_tools.exceptions import LicenseError

from ..config.licenses import LicenseConfig
from ..gomod import get_go_module_dirs
from ..licensing import combine_licenses
from .base import (
    LicenseData,
    LicenseDetector,
    LicenseDetectorNotAvailableError,
    get_manual_license_entries,
    is_unwanted_path,
    reuse_path_to_license_map,
)
from .search import find_license_files

if TYPE_CHECKING:
    from _typeshed import StrPath
    from typing_extensions import NotRequired

# Whether to pass --multiple to askalono identify
CONFIG_MULTIPLE_DEFAULT = False


class AskalonoLicenseEntry(TypedDict):
    name: str
    kind: str
    aliases: list[str]


class AskalonoLicenseContainingEntry(TypedDict):
    score: float
    license: AskalonoLicenseEntry
    line_range: list[int]


class AskalonoLicenseResult(TypedDict):
    score: float
    license: AskalonoLicenseEntry | None
    containing: list[AskalonoLicenseContainingEntry]


class AskalonoLicenseDict(TypedDict):
    path: str
    result: NotRequired[AskalonoLicenseResult]


def _remove_line(file: StrPath, key: Callable[[str], bool]) -> None:
    """
    Used to remove vendor directory from .gitignore to avoid confusing askalono
    """
    lines: list[str] = []
    with open(file, "r+", encoding="utf-8") as fp:
        for line in fp:
            if key(line):
                continue
            lines.append(line)
        fp.seek(0)
        fp.writelines(lines)
        fp.truncate()


def _filter_path(data: AskalonoLicenseDict) -> AskalonoLicenseDict:
    data["path"] = data["path"].strip("\n")
    return data


def _get_askalono_data(
    directory: StrPath, relpaths: Collection[StrPath], multiple: bool = False
) -> list[AskalonoLicenseDict]:
    stdin = "\n".join(map(str, relpaths))
    cmd = [
        "askalono",
        "--format",
        "json",
        "identify",
        "--batch",
    ]
    # --multiple seems to cause some licenses to not be detected at all, so
    # gate this behind a flag
    if multiple:
        cmd.append("--multiple")
    licenses_json = subprocess.run(
        cmd,
        input=stdin,
        check=True,
        capture_output=True,
        text=True,
        cwd=directory,
    ).stdout
    licenses = [
        _filter_path(cast(AskalonoLicenseDict, json.loads(line)))
        for line in sorted(licenses_json.splitlines())
    ]
    licenses.sort(key=lambda ld: ld.get("path", ""))
    return licenses


def _get_relative(base_dir: Path, file: str | Path) -> Path:
    file = Path(file)
    return file.relative_to(base_dir) if file.is_absolute() else file


def _get_license_name(data: AskalonoLicenseDict, check: bool) -> str | None:
    name: str | None = None
    if "result" not in data:
        pass
    elif con := data["result"]["containing"]:
        try:
            name = combine_licenses(
                *(entry["license"]["name"] for entry in con),
                validate=check,
                strict=check,
            )
        except ExpressionError as exc:  # pragma: no cover
            raise LicenseError(
                f"Failed to detect license for {data.get('path')}: {exc}"
            ) from exc
    elif data["result"]["license"]:
        name = data["result"]["license"].get("name")
    return name


def _filter_license_data(
    data: list[AskalonoLicenseDict],
    directory: Path,
) -> tuple[list[AskalonoLicenseDict], set[Path]]:

    undetected_licenses: set[Path] = set()
    results: list[AskalonoLicenseDict] = []

    for licensed in data:
        if _get_license_name(licensed, False):
            results.append(licensed)
        else:
            undetected_licenses.add(_get_relative(directory, licensed["path"]))
    return results, undetected_licenses


def _get_simplified_license_map(
    directory: Path,
    filtered_license_data: list[AskalonoLicenseDict],
    extra_license_mapping: dict[Path, str] | None = None,
) -> dict[Path, str]:
    """
    Given license data from askalono, return a simple mapping of license file
    Path to the license expression
    """
    results: dict[Path, str] = {}
    for licensed in filtered_license_data:
        license_name = _get_license_name(licensed, check=True)
        if not license_name:  # pragma: no cover
            raise RuntimeError("Should never get here after filtering the license map")
        results[_get_relative(directory, licensed["path"])] = license_name
    results.update(extra_license_mapping or {})
    return dict(sorted(results.items(), key=lambda item: item[0]))


@dataclasses.dataclass()
class AskalonoLicenseData(LicenseData):
    askalono_license_data: list[AskalonoLicenseDict]


class AskalonoLicenseDetector(LicenseDetector[AskalonoLicenseData]):
    NAME = "askalono"
    PACKAGES_NEEDED = ("askalono-cli",)

    def __init__(
        self,
        detector_config: dict[str, str],
        license_config: LicenseConfig,
        find_only: bool = False,
    ) -> None:
        self._find_only = find_only
        path: str | None = None
        if self.find_only:
            # If find_only, just set path to something
            path = "askalono"
        else:
            if path := detector_config.get("askalono_path"):
                if not Path(path).exists():
                    raise LicenseDetectorNotAvailableError(f"{path!r} does not exist!")
            else:
                path = shutil.which("askalono")
            if not path:
                raise LicenseDetectorNotAvailableError(
                    "Failed to find askalono binary!"
                )

        self.path: str = path
        self.detector_config = detector_config
        self.license_config = license_config

    def detect(self, directory: StrPath) -> AskalonoLicenseData:
        if self.find_only:
            raise ValueError(
                "This cannot be called when class was initalized with find_only=True"
            )
        gitignore = Path(directory, ".gitignore")
        if gitignore.is_file():
            _remove_line(gitignore, lambda line: line.startswith("vendor"))
        reuse_roots = get_go_module_dirs(Path(directory), relative_paths=True)
        license_file_lists = find_license_files(
            directory=directory,
            relative_paths=True,
            exclude_directories=self.license_config["exclude_directories"],
            exclude_files=self.license_config["exclude_files"],
            reuse_roots=reuse_roots,
        )
        askalono_license_data = _get_askalono_data(
            directory,
            license_file_lists["license"],
            str_to_bool(self.detector_config.get("multiple"), CONFIG_MULTIPLE_DEFAULT),
        )
        filtered_license_data, undetected = _filter_license_data(
            askalono_license_data, Path(directory)
        )
        manual_license_map, manual_unmatched = get_manual_license_entries(
            self.license_config["licenses"], directory
        )
        license_map = _get_simplified_license_map(
            Path(directory), filtered_license_data, manual_license_map
        )
        license_map |= reuse_path_to_license_map(license_file_lists["reuse"])
        # Sort
        license_map = dict(sorted(license_map.items(), key=lambda item: item[0]))
        # Remove manually specified licenses
        undetected -= set(manual_license_map)
        undetected = {
            path
            for path in undetected
            if not is_unwanted_path(
                path,
                self.license_config["exclude_directories"],
                self.license_config["exclude_files"],
            )
        }
        undetected -= set(manual_license_map)
        return AskalonoLicenseData(
            directory=Path(directory),
            license_map=license_map,
            undetected_licenses=undetected,
            unmatched_extra_licenses=manual_unmatched,
            askalono_license_data=askalono_license_data,
            # FIXME(gotmax): Change the design of LicenseData to not require full paths
            extra_license_files=[
                Path(directory, file) for file in license_file_lists["notice"]
            ],
            detector_name=self.NAME,
        )



# Copyright (C) 2025 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

"""
Test of the RPM macros
"""

from __future__ import annotations

import os
import subprocess
from collections.abc import Sequence
from pathlib import Path
from typing import NamedTuple

PARENT = Path(__file__).resolve().parent.parent.parent
# e.g. MACRO_DIR=%{buildroot}%{_rpmmacrodir} \
#      pytest
# MACRO_DIR="" MACRO_LUA_DIR="" to only use system paths
MACRO_DIR = str(os.environ.get("MACRO_DIR", PARENT / "rpm"))


class Result(NamedTuple):
    stdout: str
    stderr: str


def macros_path() -> list[str]:
    if MACRO_DIR == "":
        return []
    path = subprocess.run(
        # Don't judge. It works.
        "rpm --showrc | grep 'Macro path' | awk -F ': ' '{print $2}'",
        shell=True,
        text=True,
        check=True,
        capture_output=True,
    ).stdout.strip()
    return ["--macros", f"{path}:{MACRO_DIR}/macros.*"]


class Evaluator:
    def __init__(self) -> None:
        self.macros_path = macros_path()

    def __call__(
        self,
        exps: str | Sequence[str],
        defines: dict[str, str] | None = None,
        undefines: Sequence[str] = (),
        should_fail: bool = False,
    ) -> Result:
        cmd: list[str] = ["rpm", *self.macros_path]
        defines = defines or {}
        for name, value in defines.items():
            cmd.extend(("--define", f"{name} {value}"))
        for name in undefines:
            cmd.extend(("-E", f"%undefine {name}"))
        if isinstance(exps, str):
            cmd.extend(("-E", exps))
        else:
            for exp in exps:
                cmd.extend(("-E", exp))
        print(cmd)
        proc = subprocess.run(cmd, text=True, capture_output=True, check=False)
        if should_fail:
            assert proc.returncode != 0
        else:
            assert proc.returncode == 0, proc.stderr
        return Result(proc.stdout, proc.stderr)


evaluator = Evaluator()


def test_go_vendor_license_check_disabled():
    assert (
        evaluator(
            "%go_vendor_license_check", {"go_vendor_license_check_disable": "1"}
        ).stdout
        == "\n"
    )


def test_go_vendor_license_check():
    assert (
        evaluator("%go_vendor_license_check", {"LICENSE": "MIT"}).stdout
    ) == "go_vendor_license report expression --verify 'MIT'\n"


def test_go_vendor_license_check_args():
    assert (
        (
            evaluator(
                "%go_vendor_license_check GPL-2.0-only BSD-3-Clause", {"LICENSE": "MIT"}
            ).stdout
        )
        == "go_vendor_license report expression --verify 'GPL-2.0-only BSD-3-Clause'\n"
    )


def test_go_vendor_license_buildrequires():
    assert (
        evaluator("%go_vendor_license_buildrequires").stdout
        == "go_vendor_license generate_buildrequires\n"
    )


def test_go_vendor_license_buildrequires_disabled():
    assert (
        evaluator(
            "%go_vendor_license_buildrequires",
            {"go_vendor_license_check_disable": "1"},
        ).stdout
        == "go_vendor_license generate_buildrequires --no-check\n"
    )
